from types import SimpleNamespace
import numpy.testing as npt
from fluxpart.partition import partition_from_qc_averages


def test_partition_from_qcdata():
    """
    References
    ----------
    [PRV14] L Palatella, G Rana, D Vitale (2014), Towards a flux-
    partitioning procedure based on the direct use of high-frequency
    eddy-covariance data.  Boundary-Layer Meteorol. 153:327--337,
    doi:10.1007/s10546-014-9947-x
    """

    # April 7 "physical" example from [PRV14] Table 1
    wue = -37.158598e-3
    qcdata = SimpleNamespace(
        var_q=0.411163e-3 ** 2,
        var_c=5.182580e-6 ** 2,
        wq=0.033140e-3,
        wc=-0.472108e-6,
        corr_qc=-0.881017)
    # Results given in [PRV14]
    desired = SimpleNamespace(
        var_cp=5.230500e-6 ** 2,
        corr_cp_cr=-0.757626,
        wqt=0.012878e-3,
        wqe=0.020262e-3,
        wcp=-0.476489e-6,
        wcr=0.004381e-6)
    rsoln, fluxes = partition_from_qc_averages(qcdata, wue)
    assert_partition(rsoln, fluxes, desired)

    # April 5 "non-physical" example from [PRV14] Table 1
    wue = -24.558131e-3
    qcdata = SimpleNamespace(
        var_q=0.455994e-3 ** 2,
        var_c=4.544450e-6 ** 2,
        wq=0.062700e-3,
        wc=-0.712862e-6,
        corr_qc=-0.922292)
    # Resutls given in [PRV14]
    desired = SimpleNamespace(
        var_cp=3.951510e-6 ** 2,
        corr_cp_cr=-0.724706,
        wqt=0.025416e-3,
        wqe=0.037284e-3,
        wcp=-0.624172e-6,
        wcr=-0.088690e-6)
    rsoln, fluxes = partition_from_qc_averages(qcdata, wue)
    assert_partition(rsoln, fluxes, desired)

    # comparable to Ray Anderson peach data, 2012-06-07 1300
    # The valid solution in this case uses the '+' CO2 root
    wue = -7.060177e-3
    qcdata = SimpleNamespace(
        var_q=0.40639e-6,
        var_c=7.68505e-12,
        wq=0.1506337e-3,
        wc=-0.6254288e-6,
        corr_qc=-.9501656)
    rsoln, fluxes = partition_from_qc_averages(qcdata, wue)
    assert rsoln.validroot

    # comparable to Ray Anderson peach data, 2012-06-07 0230
    # The valid solution in this case uses the '-' CO2 root
    wue = -68.77103e-3
    qcdata = SimpleNamespace(
        var_q=0.001326586e-6,
        var_c=9.948297e-12,
        wq=0.00088955655e-3,
        wc=0.07513186e-6,
        corr_qc=0.8886955)
    rsoln, fluxes = partition_from_qc_averages(qcdata, wue)
    assert rsoln.validroot


def assert_partition(rsoln, fluxes, desired):
    if rsoln.validroot:
        npt.assert_allclose(rsoln.var_cp, desired.var_cp, atol=1e-14)
        npt.assert_allclose(rsoln.corr_cp_cr, desired.corr_cp_cr, atol=1e-2)
        npt.assert_allclose(fluxes.wqt, desired.wqt, atol=1e-7)
        npt.assert_allclose(fluxes.wqe, desired.wqe, atol=1e-7)
        npt.assert_allclose(fluxes.wcp, desired.wcp, atol=1e-9)
        npt.assert_allclose(fluxes.wcr, desired.wcr, atol=1e-10)
    else:
        assert False


if __name__ == '__main__':
    test_partition_from_qcdata()



"""Physical constants"""

from collections import namedtuple


class MolecularWeight(namedtuple('MolecularWeight', 'dryair vapor co2')):
    """
    Attributes
    ----------
    dryair, vapor, co2 : float
        kg/mol
    """


class SpecificGasConstant(namedtuple('SpecificGasConstant',
                                     'dryair vapor co2')):
    """
    Attributes
    ----------
    dryair, vapor, co2 : float
        J/K/mol == m^3 Pa/kg/K
    """


class SpecificHeatCapacity(namedtuple('SpecificHeatCapacity', 'dryair')):
    """
    Attributes
    ----------
    dryair : float
        J/K/kg
    """


GRAVITY = 9.81             # m/s^2
VON_KARMAN = 0.40          # unitless
UNIVERSAL_GAS = 8.3144598  # J/K/mol

MOLECULAR_WEIGHT = MolecularWeight(
    dryair=0.0289645,
    vapor=0.018016,
    co2=0.044010)

SPECIFIC_GAS_CONSTANT = SpecificGasConstant(
    dryair=UNIVERSAL_GAS / MOLECULAR_WEIGHT.dryair,
    vapor=UNIVERSAL_GAS / MOLECULAR_WEIGHT.vapor,
    co2=UNIVERSAL_GAS / MOLECULAR_WEIGHT.co2)

SPECIFIC_HEAT_CAPACITY = SpecificHeatCapacity(
    dryair=1004.67)



from math import exp
import pkg_resources
import numpy as np

import fluxpart
import fluxpart.partition as fp
import fluxpart.wue as wue
import fluxpart.util as util
import fluxpart.hfdata as hfdata
from fluxpart.containers import Fluxes, WUE, Result, RootSoln, HFSummary
from fluxpart.containers import QCData
from fluxpart.constants import SPECIFIC_GAS_CONSTANT as Rgas

DEFAULT_WUE_OPTIONS = {
    'ci_mod': 'const_ratio',
    'ci_mod_param': None,
    'leaf_temper': None,
    'leaf_temper_corr': 0,
    'diff_ratio': 1.6}

DEFAULT_HFD_OPTIONS = {
    'cols': (1, 2, 3, 4, 5, 6, 7),
    'unit_convert': None,
    'temper_unit': 'K',
    'bounds': None,
    'flags': None,
    'rd_tol': 0.4,
    'ad_tol': 1024,
    'ustar_tol': 0.1,
    'correcting_external': True}

DEFAULT_PART_OPTIONS = {
    'adjusting_fluxes': True}

VERSION = fluxpart.__version__

NULL_RESULT = {
    'label': "",
    'result': Result(*([""] * 5)),
    'fluxes': Fluxes(*np.full(15, np.nan)),
    'hfsummary': HFSummary(*np.full(18, np.nan)),
    'wue': WUE(*np.full(13, np.nan)),
    'rootsoln': RootSoln(*np.full(6, np.nan)),
    'qcdata': QCData(*np.full(6, np.nan))}


def flux_partition(fname, meas_wue=None, hfd_options=None, wue_options=None,
                   part_options=None, label=None):
    """Partition CO2 & H2O fluxes into stomatal & nonstomatal components.

    This is the primary user interface for :mod:`Fluxpart`. The function
    provides a full implementation of the flux partitioning algorithm: it
    reads high frequency eddy covariance data, performs necessary data
    transformations and QA/QC, analyzes water vapor and carbon dioxide
    fluxes, and partitions the fluxes into stomatal (transpiration,
    photosynthesis) and nonstomatal (evaporation, respiration)
    components using the flux variance similarity method of [SS08]_.

    The fluxpart submodule is imported in __init__ so this function can
    imported without referencing the submodule:
    ``import fluxpart.flux_partition``.

    The following notation is used in variable naming to represent
    meteorological quantities::

        u, v, w = wind velocities
        q = water vapor mass concentration
        c = carbon dioxide mass concentration
        T = air temperature
        P = total air pressure

    Parameters
    ----------
    fname : str
        Name of delimited file containing high-frequency eddy covariance
        time series data.
    hfd_options : dict, optional
        Dictionary of parameters specifying options for reading, quality
        control, and correcting high-frequency eddy covariance data. See
        ``Other parameters`` section for a listing of valid
        `hfd_options` fields. Note that `hfd_options` is optional only
        if the high-frequency data being read are all in SI units and
        the file is formatted according to the default options; see
        `hfd_options['unit_convert']` and `hfd_options['temper_unit']`
        for information about specifying and converting data units.
    meas_wue : float, optional
        Measured (or otherwise prescribed) leaf-level water use
        efficiency (kg CO2 / kg H2O). Note that by definition,
        `meas_wue` must be a negative value (< 0).
    wue_options : dict, required if `meas_wue` is not provided
        Dictionary of parameters and options used to estimate water use
        efficiency if `meas_wue` is not provided. See ``Other
        parameters`` section for a description of valid fields for
        `wue_options`. If specifying `wue_options`, it is always
        required to provide values for the 'canopy_ht', 'meas_ht', and
        'ppath' fields. Other entries are optional.
    part_options : dict, optional
        Dictionary of options for the flux partitioning procedure. See
        ``Other parameters`` section for a listing of valid options.
    label : optional
        Optional identification label/object for the data set. Could be
        a str, int, datetime object, etc.

    Returns
    -------
    dict
        {'result': :class:`~fluxpart.containers.Result`,
        'fluxes': :class:`~fluxpart.containers.Fluxes`,
        'hfsummary': :class:`~fluxpart.containers.HFSummary`,
        'wue': :class:`~fluxpart.containers.WUE`,
        'rootsoln': :class:`~fluxpart.containers.RootSoln`,
        'qcdata': :class:`~fluxpart.containers.QCData`,
        'label': `label`}

    Other Parameters
    ----------------
    hfd_options['cols'] : 7*(int,)
        7-tuple of integers indicating the column numbers of `fname`
        that contain series data for (u, v, w, q, c, T, P), in that
        order. Uses 0-based indexing. Default is (1, 2, 3, 4, 5, 6, 7)
        (thus the first column (=0) in the file is not read)
    hfd_options['unit_convert'] : dict
        Dictionary of multiplication factors required to convert any u,
        v, w, q, c, or P data not in SI units to SI units (m/s, kg/m^3,
        Pa). (Note T is not in that list). The dictionary keys are the
        variable names. For example, if all data in `fname`
        are in SI units except P and c, which are in units of kPa and
        mg/m^3, respectively, then set:
        ``hfd_options['unit_convert'] = {'P': 1e3, 'c': 1e-6}``,
        since it is necessary to multiply the kPa pressure data by 1e3
        to obtain the SI pressure unit (Pa), and the mg/m^3 CO2 data by
        1e-6 to obtain the SI concentration unit (kg/m^3).
    hfd_options['temper_unit'] : {'K', 'C'}
        The units of the temperature data T in `fname`. Default is the
        SI unit, 'K'.
    hfd_options['bounds'] : dict
        Dictionary specifying any prescribed lower and upper bounds for
        valid data. Dictionary entries have the form
        ``varname: (float, float)``, where varname is one of 'u', 'v',
        'w', 'q', 'c', 'T', or 'P', and the 2-tuple holds values for the
        lower and upper bounds: ``(lower, upper)``.  Data records are
        rejected if a variable in the record is outside the prescribed
        bounds. Default is
        ``bounds = {'c': (0, np.inf), 'q': (0, np.inf)}`` such that data
        records are rejected if c or q data are not positive values.
    hfd_options['flags'] : 2-tuple or list of 2-tuples
        Specifies that one or more columns in `fname` are used to flag
        bad data records. Each tuple is of the form (col, badval),
        where col is an int specifying the column number containing the
        flag (0-based indexing), and badval is the value of the flag
        that indicates a bad data record. Default is None.
    hfd_options['rd_tol'] : float
        Relative tolerance for rejecting the datafile. Default is
        'hfd_options['rd_tol']` = 0.4. See
        :class:`~fluxpart.hfdata.HFData`.
    hfd_options['ad_tol'] : int
        Absolute tolerance for rejecting the datafile. Default is
        `hfd_options['ad_tol']` = 1024. See
        :class:`~fluxpart.hfdata.HFData`.
    hfd_options['ustar_tol'] : float
        If the friction velocity (m/s) determined from the high
        frequency data is less than `hfd_options['ustar_tol']`, the
        partitioning analysis is aborted due to insufficient turbulence.
        Defalult is `hfd_options['ustar_tol']` = 0.1 (m/s).
    hfd_options['correcting_external'] : bool, optional
        If True (default), the water vapor and carbon dioxide series
        data are corrected for external fluctuations associated with air
        temperature and vapor density according to [WPL80]_ and [DK07]_.
    hfd_options[ other keys ]
        All other key:value pairs in `hfd_options` are passed as keyword
        arguments to numpy.genfromtxt_ (where the file is read). These
        keywords are often required to specify the details of the
        formatting of the delimited datafile.  Among the most
        commonly required are: 'delimiter', a str, int, or sequence
        that is used to separate values or define column widths (default
        is that any consecutive whitespace delimits values); and
        'skip_header', an int that specifies the number of lines to skip
        at the beginning of the file. See numpy.genfromtxt_ for a full
        description of available format options.
    wue_options['canopy_ht'] : float
        Vegetation canopy height (m).
    wue_options['meas_ht'] : float
        Eddy covariance measurement height (m).
    wue_options['ppath'] : {'C3', 'C4'}
        Photosynthetic pathway.
    wue_options['ci_mod'] : str
        Valid values: 'const_ratio', 'const_ppm', 'linear', 'sqrt'.
        See: :func:`~fluxpart.wue.water_use_efficiency`.
    wue_options['ci_mod_param'] : float or (float, float)
        Paramter values to be used with `ci_mod`.
        See: :func:`~fluxpart.wue.water_use_efficiency`.
    wue_options['leaf_temper'] : float
        Canopy leaf temperature (K). If not specified, it is assumed to
        be equal to the air temperature. See:
        :func:`~fluxpart.wue.water_use_efficiency`.
    wue_options['leaf_temper_corr'] : float
        Offset adjustment applied to canopy temperature (K).
        See: :func:`~fluxpart.wue.water_use_efficiency`.
    wue_options['diff_ratio']: float, optional
        Ratio of molecular diffusivities for water vapor and CO2.
        Default is `diff_ratio` = 1.6.
        See: :func:`~fluxpart.wue.water_use_efficiency`.
    part_options['adjusting_fluxes'] : bool
        If True (default), the final partitioned fluxes are adjusted
        proportionally such that sum of the partitioned fluxes match
        exactly the total fluxes indicated in the original data.


    .. _numpy.genfromtxt:
        http://docs.scipy.org/doc/numpy/reference/generated/numpy.genfromtxt.html

    """

    null_result = NULL_RESULT
    # py3.5 dictionary merge/update
    hfd_options = {**DEFAULT_HFD_OPTIONS, **(hfd_options or {})}
    wue_options = {**DEFAULT_WUE_OPTIONS, **(wue_options or {})}
    part_options = {**DEFAULT_PART_OPTIONS, **(part_options or {})}

    usecols = np.array(hfd_options.pop('cols'), dtype=int).reshape(7,)

    converters = None
    unit_convert = hfd_options.pop('unit_convert')
    if unit_convert:
        converters = {
            k: _converter_func(float(v), 0.) for k, v in unit_convert.items()}

    temper_unit = hfd_options.pop('temper_unit')
    if temper_unit.upper() == 'C' or temper_unit.upper() == 'CELSIUS':
        converters = converters or {}
        converters['T'] = _converter_func(1., 273.15)

    correcting_external = hfd_options.pop('correcting_external')
    ustar_tol = hfd_options.pop('ustar_tol')

    # read high frequency data
    try:
        hfdat = hfdata.HFData(fname, cols=usecols, converters=converters,
                              **hfd_options)
    except hfdata.Error as err:
        mssg = 'High frequency data error: ' + err.args[0]
        result = Result(version=VERSION, dataread=False,
                        attempt_partition=False, valid_partition=False,
                        mssg=mssg)
        return {**null_result,
                'label': label,
                'result': result}

    # preliminary data processing and analysis
    hfdat.truncate()
    if correcting_external:
        hfdat.qc_correct()
    hfsum = hfdat.summarize()

    # exit if friction velocity is too low (lack of turbulence)
    if hfsum.ustar < ustar_tol:
        mssg = ('ustar = {:.4} is less than ustar_tol = {:.4}'.
                format(hfsum.ustar, ustar_tol))
        result = Result(version=VERSION, dataread=True, attempt_partition=False,
                        valid_partition=False, mssg=mssg)
        return {**null_result,
                'label': label,
                'result': result,
                'hfsummary': hfsum}

    # exit if atmospheric vapor pressure deficit is <= 0
    Tr = 1 - 373.15 / hfsum.T
    esat = 101325. * (
        exp(13.3185 * Tr - 1.9760 * Tr**2 - 0.6445 * Tr**3 - 0.1299 * Tr**4))
    vpd = esat - hfsum.rho_vapor * Rgas.vapor * hfsum.T
    if vpd <= 0:
        mssg = 'Vapor pressure deficit {}'.format(vpd)
        result = Result(version=VERSION, dataread=True, attempt_partition=False,
                        valid_partition=False, mssg=mssg)
        return {**null_result,
                'label': label,
                'result': result,
                'hfsummary': hfsum}

    # exit if water vapor flux is downward (negative)
    if hfsum.cov_w_q <= 0:
        mssg = ('cov(w,q) = {:.4} <= 0 is incompatible with partitioning '
                'algorithm'.format(hfsum.cov_w_q))
        result = Result(version=VERSION, dataread=True, attempt_partition=False,
                        valid_partition=False, mssg=mssg)
        return {**null_result,
                'label': label,
                'result': result,
                'hfsummary': hfsum}

    # get or calculate water use efficiency
    if meas_wue:
        leaf_wue = WUE(float(meas_wue), *np.full(12, np.nan))
    else:
        try:
            leaf_wue = wue.water_use_efficiency(hfsum, **wue_options)
        except wue.Error as err:
            result = Result(version=VERSION, dataread=True, attempt_partition=False,
                            valid_partition=False, mssg=err.args[0])
            return {**null_result,
                    'label': label,
                    'result': result,
                    'hfsummary': hfsum}

    # compute partitioned fluxes
    adjusting_fluxes = part_options['adjusting_fluxes']
    pout = fp.partition_from_wqc_series(hfdat['w'], hfdat['q'], hfdat['c'],
                                        leaf_wue.wue, adjusting_fluxes)

    # collect results and return
    result = Result(version=VERSION, dataread=True,
                    attempt_partition=True,
                    valid_partition=pout['valid_partition'],
                    mssg=pout['partmssg'])

    if pout['valid_partition']:
        fluxes = Fluxes(
            *pout['fluxcomps'],
            LE=util.qflux_mass_to_heat(pout['fluxcomps'].wq, hfsum.T),
            LEt=util.qflux_mass_to_heat(pout['fluxcomps'].wqt, hfsum.T),
            LEe=util.qflux_mass_to_heat(pout['fluxcomps'].wqe, hfsum.T),
            Fq_mol=util.qflux_mass_to_mol(pout['fluxcomps'].wq),
            Fqt_mol=util.qflux_mass_to_mol(pout['fluxcomps'].wqt),
            Fqe_mol=util.qflux_mass_to_mol(pout['fluxcomps'].wqe),
            Fc_mol=util.cflux_mass_to_mol(pout['fluxcomps'].wc),
            Fcp_mol=util.cflux_mass_to_mol(pout['fluxcomps'].wcp),
            Fcr_mol=util.cflux_mass_to_mol(pout['fluxcomps'].wcr))
    else:
        fluxes = Fluxes(*np.full(15, np.nan))

    return {'label': label,
            'result': result,
            'fluxes': fluxes,
            'hfsummary': hfsum,
            'wue': leaf_wue,
            'rootsoln': pout['rootsoln'],
            'qcdata': pout['qcdata']}


def _converter_func(slope, intercept):
    """Return a function for linear transform of data when reading file.
    """
    def func(stringval):
        try:
            return slope * float(stringval.strip()) + intercept
        except ValueError:
            return np.nan
    return func


if __name__ == "__main__":
    pass



"""Data strucure for high-frequency eddy covarinace time series."""

import math
import numpy as np

import fluxpart.util as util
from fluxpart.containers import HFSummary
from fluxpart.constants import MOLECULAR_WEIGHT as MW
from fluxpart.constants import SPECIFIC_HEAT_CAPACITY as CP
from fluxpart.constants import SPECIFIC_GAS_CONSTANT as GC


class Error(Exception):
    pass


class HFData:
    """High-frequency eddy covariance data.

    The following notation is used in variable naming and documentation
    to represent meteorological quantities (SI units)::

        u, v, w = wind velocities (m/s)
        q = water vapor mass concentration (kg/m^3)
        c = carbon dioxide mass concentration (kg/m^3)
        T = air temperature (K)
        P = total air pressure (Pa)

    Parameters
    ----------
    fname : str
        Name of delimited file containing high-frequency eddy covariance
        time series data.
    cols : 7*(int,)
        7-tuple of integers indicating the column numbers of `fname`
        that contain series data for (u, v, w, q, c, T, P), in that
        order. Uses 0-based indexing.
    converters : dict, optional
        Dictionary of functions used to convert any non-SI data to SI
        units.  Dict keys are 'u', 'v', 'w', 'q', 'c', 'T', or 'P', and
        the funcs take string single argument, e.g.
        ``converters={'P': lambda s: 1e3 * float(s)}``.
    bounds : dict, optional
        Dictionary specifying any prescribed lower and upper bounds for
        legal data. Dict entries have the form
        ``varname: (float, float)``, where varname is one of 'u', 'v',
        'w', 'q', 'c', 'T', or 'P', and the 2-tuple holds values for the
        lower and upper bounds: ``(lower, upper)``.  Data records are
        rejected if a variable in the record is outside the prescribed
        bounds. Default is
        ``bounds = {'c': (0, np.inf), 'q': (0, np.inf)}`` such that data
        records are rejected if c or q data are not positive values.
    flags : 2-tuple or list of 2-tuples, optional
        Specifies that one or more columns in `fname` are used to flag
        bad data records. Each tuple is of the form (col, badval),
        where col is an int specifying the column number containing the
        flag (0-based indexing), and badval is the value of the flag
        that indicates a bad data record.
    rd_tol : float, optional
        Relative tolerance for rejecting the datafile. Default is
        `rd_tol` = 0.4.  See `ad_tol` for explanation.
    ad_tol : int, optional
        Absolute tolerance for rejecting the datafile. Defaults is
        `ad_tol` = 1024. If the datafile contains bad records (not
        readable, out-of-bounds, or flagged data), the partitioning
        analysis is performed using the longest stretch of consecutive
        good data records found, unless that stretch is too short,
        in which case the analysis is aborted. The criteria for
        judging 'too short' can be specified in both relative and
        absolute terms: the datafile is rejected if the good stretch
        is a fraction of the total data that is less than `rd_tol`,
        and/or is less than `ad_tol` records long.
    **kwargs
        Keyword arguments passed to numpy.genfromtxt_ to specify
        formatting of the delimited datafile. See numpy.genfromtxt_
        for a full description of available options.


    .. _numpy.genfromtxt:
        http://docs.scipy.org/doc/numpy/reference/generated/numpy.genfromtxt.html

    Attributes
    ----------
    data_table : structured array
        Table of high frequency data series with columns 'u', 'v', 'w',
        'q', 'c', 'T', and 'P'
    """

    def __init__(self, fname, cols, converters=None, flags=None, bounds=None,
                 rd_tol=0.5, ad_tol=1024, **kwargs):
        """Read high frequency eddy covarince data, apply some QA/QC."""

        var_names = ['u', 'v', 'w', 'q', 'c', 'T', 'P']
        self.names = var_names.copy()
        all_names = var_names.copy()
        self._qc_already_corrected = False
        usecols = np.array(cols, dtype=int).reshape(7,)

        # make sure no duplicate kws
        disallowed_kws = ('usecols', 'names', 'dtype', 'converters',
                          'filling_values')
        for kw in disallowed_kws:
            delete_if_present = kwargs.pop(kw, False)

        var_dtype = list(zip(var_names, 7 * (float, )))
        all_dtype = var_dtype.copy()

        # add flag columns to data table
        if flags:
            if isinstance(flags, list):
                fcols, goodvals = list(zip(*flags))
            else:
                fcols, goodvals = [flags[0], ], [flags[1], ]
            flag_names = list('flg' + str(i) for i in range(len(fcols)))

            usecols = np.append(usecols, np.array(fcols, dtype=int))
            all_dtype += list(zip(flag_names, [type(v) for v in goodvals]))
            all_names += flag_names

            # for later convenience
            flag_dict = dict(zip(flag_names, goodvals))

        try:
            data = np.genfromtxt(fname, usecols=usecols, dtype=all_dtype,
                                 names=all_names, converters=converters,
                                 **kwargs)
        except TypeError as err:
            raise Error(err.args[0])

        # 1D bool mask is True if any `data` field is missing (=nan)
        mask = np.isnan(data[var_names].view(float).reshape(-1, 7)).any(axis=1)

        # Also mask records with out-of-bounds or flagged data
        dbounds = {'c': (0, np.inf), 'q': (0, np.inf)}
        if bounds:
            dbounds.update(bounds)
        for var, (low, high) in dbounds.items():
            mask[(data[var] <= low) | (data[var] >= high)] = True
        if flags:
            for flgname, goodval in flag_dict.items():
                mask[data[flgname] != goodval] = True

        # Find longest span of valid (unmasked) data
        dummy_ma = np.ma.array(np.zeros([data.shape[0], ]), mask=mask)
        unmasked_slices = np.ma.clump_unmasked(dummy_ma) or [slice(0, 0), ]
        max_indx = np.argmax([s.stop - s.start for s in unmasked_slices])
        max_slice = unmasked_slices[max_indx]
        len_max_slice = max_slice.stop - max_slice.start

        # verify sufficient data length
        data_frac = len_max_slice / data.shape[0]
        if data_frac < rd_tol or len_max_slice < ad_tol:
            self.data_table = None
            raise Error(
                'HF Data read but rejected because the longest continuous run '
                'of valid data was too short on a relative (length data / '
                'total length) = {:.4} < dtol = {:.4}) AND/OR absolute basis '
                '(data length = {} < {})'
                ''.format(data_frac, rd_tol, len_max_slice, ad_tol))

        self.data_table = data[var_names][max_slice].copy()
        return

    def __getitem__(self, name):
        """Column-wise get without specifying data_table attribute"""
        return self.data_table[name]

    def __setitem__(self, name, value):
        """Column-wise set without specifying data_table attribute"""
        self.data_table[name] = value

    def truncate(self):
        """Truncate length of data_table view to largest possible power of 2."""
        truncate_len = 2 ** int(np.log2(self.data_table.shape[0]))
        self.data_table = self.data_table[:truncate_len]

    def qc_correct(self):
        """Adjust q and c data series to correct for external effects.

        Water vapor and carbon dioxide series data in the data_table are
        corrected for external fluctuations associated with air
        temperature and vapor density. See: [WPL80]_ and [DK07]_.

        """

        if self._qc_already_corrected:
            return
        ave_vapor = self['q'].mean()
        ave_co2 = self['c'].mean()
        ave_T = self['T'].mean()
        dev_vapor = self['q'] - ave_vapor
        dev_T = self['T'] - ave_T

        Pdryair = self['P'].mean() - ave_vapor * GC.vapor * ave_T
        rho_totair = ave_vapor + Pdryair / GC.dryair / ave_T

        specific_vapor = ave_vapor / rho_totair
        specific_co2 = ave_co2 / rho_totair
        mu = MW.dryair / MW.vapor
        muq = mu * specific_vapor
        muc = mu * specific_co2

        self['q'] += muq * dev_vapor + (1 + muq) * ave_vapor * dev_T / ave_T
        self['c'] += muc * dev_vapor + (1 + muq) * ave_co2 * dev_T / ave_T
        self._qc_already_corrected = True
        return

    def summarize(self):
        """Return statisitical summary of high frequency data table.

        Returns
        -------
        namedtuple
            :class:`~fluxpart.containers.HFSummary`

        """

        hfs = util.stats2(self.data_table, self.names)
        Pvap = hfs.ave_q * GC.vapor * hfs.ave_T
        rho_dryair = (hfs.ave_P - Pvap) / GC.dryair / hfs.ave_T
        rho_totair = rho_dryair + hfs.ave_q
        Cp = CP.dryair * (1 + 0.84 * hfs.ave_q / rho_totair)

        return HFSummary(
            T=hfs.ave_T,
            P=hfs.ave_P,
            Pvap=Pvap,
            ustar=(hfs.cov_w_u**2 + hfs.cov_w_v**2)**0.25,
            wind_w=hfs.ave_w,
            var_w=hfs.var_w,
            rho_vapor=hfs.ave_q,
            var_vapor=hfs.var_q,
            rho_co2=hfs.ave_c,
            var_co2=hfs.var_c,
            corr_q_c=hfs.cov_q_c / math.sqrt(hfs.var_q * hfs.var_c),
            H=rho_totair * Cp * hfs.cov_w_T,
            cov_w_q=hfs.cov_w_q,
            cov_w_c=hfs.cov_w_c,
            rho_dryair=rho_dryair,
            rho_totair=rho_totair,
            cov_w_T=hfs.cov_w_T,
            N=self.data_table.shape[0])


if __name__ == "__main__":
    pass



"""Namedtuples used to group data and results."""

from collections import namedtuple
import numpy as np


class FluxComponents(namedtuple('FluxComponents', 'wq wqt wqe wc wcp wcr')):
    """Vapor and CO2 flux components.

    Attributes
    ----------
    wq, wqt, wqe : float
        Total, transpiration, and evaporation H2O fluxes, kg/m^2/s
    wc, wcp, wcr : float
        Total, photosynthesis, and respiration CO2 fluxes, kg/m^2/s

    """

    __slots__ = ()

    def __str__(self):
        # Print with common mass units instead of SI
        wqs = [1e3 * self.wq, 1e3 * self.wqt, 1e3 * self.wqe]
        wcs = [1e6 * self.wc, 1e6 * self.wcp, 1e6 * self.wcr]
        return ('FluxComponents(\n'
                '    wq = {:.4} g/m^2/s,\n'
                '    wqt = {:.4} g/m^2/s,\n'
                '    wqe = {:.4} g/m^2/s,\n'
                '    wc = {:.4} mg/m^2/s,\n'
                '    wcp = {:.4} mg/m^2/s,\n'
                '    wcr = {:.4} mg/m^2/s)'
                ''.format(*(wqs + wcs)))


class Fluxes(namedtuple('Fluxes', 'Fq Fqt Fqe Fc Fcp Fcr LE LEt LEe '
                        'Fq_mol Fqt_mol Fqe_mol Fc_mol Fcp_mol Fcr_mol')):

    """Water vapor and CO2 fluxes.

    Attributes
    ----------
    Fq, Fqt, Fqe : float
        Total, transpiration, and evaporation H2O fluxes, kg/m^2/s.
    Fc, Fcp, Fcr : float
        Total, photosynthesis, and respiration CO2 fluxes, kg/m^2/s.
    LE, LEt, LEe : float
        Same vapor fluxes expressed as latent heat, W/m^2.
    Fq_mol, Fqt_mol, Fqe_mol : float
        Same vapor fluxes expressed as mol/m^2/s.
    Fc_mol, Fcp_mol, Fcr_mol : float
        Same CO2 fluxes expressed as mol/m^2/s.

    """

    __slots__ = ()

    def __str__(self):
        # For some fields, print common units instead of SI
        row = "{:35}{}"
        return (
            'Fluxes(\n' +
            row.format(
                '    Fq = {:.4} g/m^2/s,'.format(1e3 * self.Fq),
                'Fc = {:.4} mg/m^2/s,\n'.format(1e6 * self.Fc)) +
            row.format(
                '    Fqt = {:.4} g/m^2/s,'.format(1e3 * self.Fqt),
                'Fcp = {:.4} mg/m^2/s,\n'.format(1e6 * self.Fcp)) +
            row.format(
                '    Fqe = {:.4} g/m^2/s,'.format(1e3 * self.Fqe),
                'Fcr = {:.4} mg/m^2/s,\n\n'.format(1e6 * self.Fcr)) +
            row.format(
                '    Fq_mol = {:.4} mmol/m^2/s,'.format(1e3 * self.Fq_mol),
                'Fc_mol = {:.4} umol/m^2/s,\n'.format(1e6 * self.Fc_mol)) +
            row.format(
                '    Fqt_mol = {:.4} mmol/m^2/s,'.format(1e3 * self.Fqt_mol),
                'Fcp_mol = {:.4} umol/m^2/s,\n'.format(1e6 * self.Fcp_mol)) +
            row.format(
                '    Fqe_mol = {:.4} mmol/m^2/s,'.format(1e3 * self.Fqe_mol),
                'Fcr_mol = {:.4} umol/m^2/s,\n\n'.format(1e6 * self.Fcr_mol)) +
            '    LE  = {:.4} W/m^2,\n'.format(self.LE) +
            '    LEt = {:.4} W/m^2,\n'.format(self.LEt) +
            '    LEe = {:.4} W/m^2)'.format(self.LEe))


class HFSummary(namedtuple('HFSummary', 'T P Pvap ustar wind_w var_w '
                           'rho_vapor rho_co2 var_vapor var_co2 corr_q_c, '
                           'cov_w_q cov_w_c H rho_dryair rho_totair '
                           'cov_w_T N')):

    """Summary of high frequency eddy covariance data.

    Attributes
    ----------
    T : float
        Mean air temperature, K.
    P, Pvap : float
        Mean total atmospheric (`P`) and vapor (`Pvap`) pressure, Pa.
    ustar : float
        Mean friction velocity, m/s.
    wind_w : float
        Mean vertical wind velocity, m/s.
    var_w : float
        Variance of vertical wind velocity, (m/s)^2.
    rho_vapor, rho_co2 : float
        Mean H2O vapor and CO2 concentrations, kg/m^3.
    var_vapor, var_co2 : float
        Variance of H2O vapor and CO2 concentrations, (kg/m^3)^2.
    corr_q_c : float
        Correlation coefficient for H2O and CO2 concentrations
    cov_w_q, cov_w_c : float
        Covariance of vertical wind velocity (w) with water vapor (q)
        and CO2 (c) mass densities, kg/m^2/s.
    H : float
        Sensible heat flux, W/m^2.
    rho_dryair, rho_totair : float
        Dry and moist air densities, kg/m^3.
    cov_w_T : float
        Covariance of temperature and vertical wind velocity, K m/s.
    N : int
        Length of data series.

    """

    __slots__ = ()

    def __str__(self):

        # For some fields, print common units instead of SI
        dum = self._replace(
            T=self.T - 273.15,
            P=1e-3 * self.P,
            Pvap=1e-3 * self.Pvap,
            rho_vapor=1e3 * self.rho_vapor,
            rho_co2=1e6 * self.rho_co2,
            var_vapor=1e6 * self.var_vapor,
            var_co2=1e12 * self.var_co2,
            cov_w_q=1e3 * self.cov_w_q,
            cov_w_c=1e6 * self.cov_w_c)

        return ('HFSummary(\n'
                '    T = {:.4} C,\n'
                '    P = {:.4} kPa,\n'
                '    Pvap = {:.4} kPa,\n'
                '    ustar = {:.4} m/s,\n'
                '    wind_w = {:.4} m/s,\n'
                '    var_w = {:.4} (m/s)^2,\n'
                '    rho_vapor = {:.4} g/m^3,\n'
                '    rho_co2 = {:.4} mg/m^3,\n'
                '    var_vapor = {:.4} (g/m^3)^2,\n'
                '    var_co2 = {:.4} (mg/m^3)^2,\n'
                '    corr_q_c = {:.4},\n'
                '    cov_w_q = {:.4} g/m^2/s,\n'
                '    H = {:.4} W/m^2,\n'
                '    cov_w_c = {:.4} mg/m^2/s,\n'
                '    rho_dryair = {:.4} kg/m^3,\n'
                '    rho_totair = {:.4} kg/m^3,\n'
                '    cov_w_T = {:.4} C m/s,\n'
                '    N = {})'
                ''.format(*dum))


class RootSoln(namedtuple('RootSoln', 'corr_cp_cr var_cp sig_cr co2soln_id '
                           'validroot validmssg')):
    """Results from calcuating the root (corr_cp_cr, var_cp).

    Attributes
    ----------
    corr_cp_cr : float
        Correlation coefficient for CO2 concentrations connected with
        photosynthesis (cp) and respiration (cr).
    var_cp : float
        Variance of CO2 concentration connected with photosynthesis,
        (kg/m^3)^2.
    sig_cr : float
        Standard deviation of CO2 concentration connected with
        respiration, kg/m^3.
    co2soln_id : {0 or 1}
        Indicates the solution used for the quadratic Eq. 13b of
        [SS08]_. Equal to 1 for the '+' root, equal to 0 for the '-'
        root.
    validroot : bool
        Indicates whether the obtained root (`corr_cp_cr`, `var_cp`) is
        physically plausible.
    validmssg : str
        Possibly informative message if `validroot` = False.

    """

    __slots__ = ()

    def __str__(self):

        # For some fields, print common units instead of SI
        dum = self._replace(
            var_cp=1e12 * self.var_cp,
            sig_cr=1e6 * self.sig_cr)

        return ('RootSoln(\n'
                '    corr_cp_cr = {0:.4},\n'
                '    var_cp = {1:.4} (mg/m^3)^2,\n'
                '    sig_cr = {2:.4} mg/m^3,\n'
                '    co2soln_id = {3},\n'
                '    validroot = {4},\n'
                '    validmssg = {5})'
                ''.format(*dum))


class QCData(namedtuple('QCData', 'var_q var_c corr_qc wq wc wave_lvl')):

    """Summary stats for water vapor and CO2 series data.

    Attributes
    ----------
    var_q, var_c : float
        Variance of vapor (q) and CO2 (c) concentrations, (kg/m^3)^2.
    corr_qc : float
        Correlation coefficient for vapor and CO2 concentrations.
    wq, wc : float
        Mean vapor (`wq`) and CO2 (`wc`) fluxes, kg/m^2/s.
    wave_lvl : (int, int)
        2-tuple indicating the level of filtering applied (number of
        components removed from the series data). The second int is the
        maximum possible wavelet decompostion level given the length of
        the data. The first is the number of components remaining in the
        data.  So when the first number is equal to the second, no
        components have been removed (no filtering applied). When the
        first number is 1, the maximum level of filtering was applied.

    """

    __slots__ = ()

    def __str__(self):

        # For some fields, print common units instead of SI
        dum = self._replace(
            var_q=1e6 * self.var_q,
            var_c=1e12 * self.var_c,
            wq=1e3 * self.wq,
            wc=1e6 * self.wc)

        return ('QCData(\n'
                '    var_q = {:.4} (g/m^3)^2,\n'
                '    var_c = {:.4} (mg/m^3)^2,\n'
                '    corr_qc = {:.4},\n'
                '    wq = {:.4} g/m^2/s,\n'
                '    wc = {:.4} mg/m^2/s,\n'
                '    wave_lvl = {})'
                ''.format(*dum))


class Result(namedtuple('Result', 
                        'version dataread attempt_partition valid_partition '
                        'mssg')):
    """Overall outcome of partitioning.

    Attributes
    ----------
    version : str
        Fluxpart version
    dataread, attempt_partition, valid_partition : bool
        Indicates success or failure in reading high frequency data,
        attempting and obtaining a valid partioning solution.
    mssg : str
        Possibly informative message if `dataread` or `valid_partition`
        are False

    """

    __slots__ = ()

    def __str__(self):
        return ('Result(\n'
                '    version = {},\n'
                '    dataread = {},\n'
                '    attempt_partition = {},\n'
                '    valid_partition = {},\n'
                '    mssg = {})'
                ''.format(*self))


class WUE(namedtuple('WUE',
                     'wue inter_h2o inter_co2 ambient_h2o ambient_co2 vpd '
                     'ci_mod ci_mod_param leaf_temper ppath meas_ht '
                     'canopy_ht diff_ratio')):
    """Summary of leaf-level water use efficiency calculation.

    Attributes
    ----------
    wue : float
        Leaf-level water use efficiency, kg CO2 / kg H2O.
    inter_h2o : float
        Concentration of intercellular water vapor, kg/m^3
    inter_co2 : float
        Concentrations of intercellular CO2, kg/m^3.
    ambient_h2o : float
        Concentrations of ambient atmospheric water vapor, kg/m^3.
    ambient_co2 : float
        Concentration of ambient CO2, kg/m^3.
    vpd : float
        Atmospheric vapor pressure deficit, Pa.
    ci_mod : str
        The name of the model used to estimate the intercellular CO2
        concentration.
    ci_mod_param : float or 2-tuple of floats
        Specific paramter values used with `ci_mod`.
    leaf_temper : float
    ppath : {'C3' or 'C4'}
        Photosynthetic pathway.
    meas_ht, canopy_ht : float
        Eddy covariance measument height and plant canopy height, m.

    """

    __slots__ = ()

    def __str__(self):

        # For some fields, print common units instead of SI
        dum = self._replace(
            wue=1e3 * self.wue,
            inter_h2o=1e3 * self.inter_h2o,
            inter_co2=1e6 * self.inter_co2,
            ambient_h2o=1e3 * self.ambient_h2o,
            ambient_co2=1e6 * self.ambient_co2,
            vpd=1e-3 * self.vpd,
            leaf_temper=self.leaf_temper - 273.15)

        return ('WUE(\n'
                '    wue = {:.4} mg/g,\n'
                '    inter_h2o = {:.4} g/m^3,\n'
                '    inter_co2 = {:.4} mg/m^3,\n'
                '    ambient_h2o = {:.4} g/m^3,\n'
                '    ambient_co2 = {:.4} mg/m^3,\n'
                '    vpd = {:.4} kPa,\n'
                '    ci_mod = {},\n'
                '    ci_mod_param = {},\n'
                '    leaf_temper = {:.4} C,\n'
                '    ppath = {},\n'
                '    meas_ht = {} m,\n'
                '    canopy_ht = {} m,\n'
                '    diff_ratio = {})'
                ''.format(*dum))



from collections import namedtuple
from itertools import permutations

import numpy as np
import pywt

from fluxpart.constants import MOLECULAR_WEIGHT as MW    # kg/mol


def stats2(sarray, names=None):
    """Calculate means and (co)variances for structured array data."""

    if names is None:
        names = sarray.dtype.names
    nvar = len(names)
    data = tuple(sarray[name] for name in names)
    cov = np.cov(data)
    nondiag_cov = list(cov[i, j] for i, j in permutations(range(nvar), 2))

    names_ave = list('ave_' + name for name in names)
    names_var = list('var_' + name for name in names)
    names_cov = list(
        'cov_' + n1 + "_" + n2 for n1, n2 in permutations(names, 2))

    out = dict(zip(names_ave, np.mean(data, axis=1)))
    out.update(zip(names_var, cov.diagonal()))
    out.update(zip(names_cov, nondiag_cov))

    NamedStats = namedtuple('Stats2', names_ave + names_var + names_cov)
    return NamedStats(**out)


def qflux_mass_to_heat(massflux, Tk):                # kg/m^2/s, K
    Lv = 2.5008e6 - 2366.8 * (Tk - 273.15)           # J/kg
    return massflux * Lv                             # W/m^2


def cflux_mass_to_mol(massflux):                     # kg/m^2/s
    return 1. / MW.co2 * massflux                    # mol/m^2/s


def qflux_mass_to_mol(massflux):                     # kg/m^2/s
    return 1. / MW.vapor * massflux                  # mol/m^2/s


def progressive_lowcut_series(series):
    """Progressively remove low-frequency components of 1D series.

    Yields sequence in which the low-frequency (large-scale) components
    of `series` are progressively removed.  The sequence is obtained
    from reconstrution of a multilevel discrete haar wavelet
    decompositon of `series`.

    N.B.: The length of `series` is assumed to be a power of 2
    (does not check!)

    Parameters
    ----------
    series : array_like
        1D data series with a length that is a power of 2

    Yields
    -------
    lowcut_series : array
        Sequence of progressively lowcut filtered data `series`. The
        yielded series have the same length as `series`.

    Notes
    -----
    After an N level discrete wavelet decomposition, a data series S can
    be reconstructed in terms of wavelet 'approximations' (A) and
    'details' (D)::

        S = A(N) + D(N) + D(N-1) ... D(2) + D(1)
          = A(N-1) + D(N-1) + ... D(2) + D(1)        [A(N-1) = A(N) + D(N)]
            ...
          = A(1) + D(1)                              [(A(1) = A(2) + D(2)]

    where A(N) represents the 'lowest' level approximation [e.g., for
    the haar wavelet and a complete decomposition of dyadic length data,
    A(N) is equal to mean(S)]

    The sequence returned by this function is::

        S - A(N),
        S - A(N-1),
        S - A(N-2),
        ...,
        S - A(1)

    This sequence is computed by the equivalent::

        S - A(N),
        S - A(N) - D(N),
        S - A(N) - D(N) - D(N-1),
        ...,
        S - A(N) - D(N) - D(N-1) - ... - D(2),

    i.e. the details are removed in sequence::

        lowcut = S - A(N)
        for j = N to 2
            lowcut = lowcut - D(j)

    """

    series_data = np.asarray(series)
    wavelet = pywt.Wavelet('haar')
    nlevel = pywt.dwt_max_level(series_data.size, wavelet.dec_len)
    decomp_coef = pywt.wavedec(series_data, wavelet=wavelet, level=nlevel)
    cAn, cD = decomp_coef[0], decomp_coef[1:]
    lowcut_series = series_data - pywt.upcoef('a', cAn, wavelet, level=nlevel)
    yield lowcut_series
    for j, cDj in enumerate(cD[:-1]):
        Dj = pywt.upcoef('d', cDj, wavelet, level=nlevel - j)
        lowcut_series -= Dj
        yield lowcut_series


if __name__ == "__main__":
    pass



import os
from types import SimpleNamespace
import numpy.testing as npt
from fluxpart import flux_partition

TESTDIR = os.path.dirname(os.path.realpath(__file__))


def test_flux_partition():
    """Integrated test of highest level flux_partition function.

    In test below, the "desired" results are not exact or even
    necessarily correct. They were obtained with an independent (matlab)
    code, and there should be reasonable agreement.
    """

    wue_data = {'meas_ht': 7.11, 'canopy_ht': 4.42, 'ppath': 'C3',
                'ci_mod': 'const_ppm', 'diff_ratio': 1/0.7}

    # soln exists for this data without any wavelet filtering
    fname = os.path.join(TESTDIR,
                         'data/TOA5_6843.ts_Above_2012_06_07_1300.dat')

    matlab_fluxes = SimpleNamespace(
        Fcp=-1.02807378762793e-6,
        Fcr=0.402683101177712e-6,
        Fqe=0.00500869036088203e-3,
        Fqt=0.145615860044424e-3,
        Fcp_mol=-23.3600042633021e-6,
        Fcr_mol=9.14980916104776e-6,
        LEe=12.1870591763138,
        LEt=354.310004313924)

    hfd = {
        'cols': (2, 3, 4, 6, 5, 7, 8),
        'delimiter': ",",
        'skip_header': 4,
        'unit_convert': {'q': 1e-3, 'c': 1e-6, 'P': 1e3},
        'temper_unit': 'C'}

    result = flux_partition(
        fname,
        wue_options=wue_data,
        hfd_options=hfd)

    npt.assert_allclose(result['rootsoln'].var_cp, 18.9272e-12, atol=1e-12)
    assert_flux_components(result['fluxes'], matlab_fluxes)

    # soln is obtained after some wavelet filtering
    fname = os.path.join(TESTDIR,
                         'data/TOA5_6843.ts_Above_2012_06_07_1245.dat')

    matlab_fluxes = SimpleNamespace(
        Fcp=-0.866856083109642e-6,
        Fcr=0.353428894620522e-6,
        Fqe=0.0124697200158396e-3,
        Fqt=0.117438136138301e-3,
        Fcp_mol=-23.1074540435422e-6,
        Fcr_mol=10.6590633820467e-6,
        LEe=35.6007693518818,
        LEt=335.283229492226)

    result = flux_partition(
        fname,
        wue_options=wue_data,
        hfd_options=hfd)

    npt.assert_allclose(result['rootsoln'].var_cp, 15.2944e-12, atol=1e-12)
    assert_flux_components(result['fluxes'], matlab_fluxes)


def assert_flux_components(calc, desired):
    npt.assert_allclose(calc.Fcp, desired.Fcp, atol=0.5e-6)
    npt.assert_allclose(calc.Fcr, desired.Fcr, atol=0.1e-6)
    npt.assert_allclose(calc.Fqe, desired.Fqe, atol=0.01e-3)
    npt.assert_allclose(calc.Fqt, desired.Fqt, atol=0.05e-3)
    npt.assert_allclose(calc.Fcp_mol, desired.Fcp_mol, atol=10e-6)
    npt.assert_allclose(calc.Fcr_mol, desired.Fcr_mol, atol=10e-6)
    npt.assert_allclose(calc.LEe, desired.LEe, atol=10)
    npt.assert_allclose(calc.LEt, desired.LEt, atol=50)


if __name__ == "__main__":
    test_flux_partition()



import os, io
import numpy as np
import numpy.testing as npt
from fluxpart.hfdata import HFData
from fluxpart.fluxpart import _converter_func

TESTDIR = os.path.dirname(os.path.realpath(__file__))


def test_hfdata_from_txt():
    cols = (2, 3, 4, 6, 5, 7, 8)
    fname = os.path.join(TESTDIR,
                         'data/TOA5_6843.ts_Above_2012_06_07_1300.dat')

    data = HFData(
        fname,
        cols,
        converters={
            'T': _converter_func(1, 273.15),
            'q': _converter_func(1e-3, 0),
            'c': _converter_func(1e-6, 0),
            'P': _converter_func(1e3, 0)},
        flags=(9, 0),
        delimiter=",",
        skip_header=4)

    npt.assert_allclose(data['u'][0], 0.468)
    npt.assert_allclose(data['v'][0], -0.9077501)
    npt.assert_allclose(data['w'][0], 0.1785)
    npt.assert_allclose(data['c'][0], 659.7584e-6)
    npt.assert_allclose(data['q'][0], 9.530561e-3)
    npt.assert_allclose(data['T'][0], 28.52527 + 273.15)
    npt.assert_allclose(data['P'][0], 100.1938e3)
    npt.assert_allclose(data['u'][-1], 1.3675)
    npt.assert_allclose(data['v'][-1], -0.75475)
    npt.assert_allclose(data['w'][-1], -0.1775)
    npt.assert_allclose(data['c'][-1], 658.2624e-6)
    npt.assert_allclose(data['q'][-1], 9.404386e-3)
    npt.assert_allclose(data['T'][-1], 28.35199 + 273.15)
    npt.assert_allclose(data['P'][-1], 100.1938e3)

    npt.assert_allclose(data['u'].mean(), 1.43621, atol=1e-4)
    npt.assert_allclose(data['v'].mean(), -0.634818, atol=1e-4)
    npt.assert_allclose(data['w'].mean(), 0.0619483, atol=1e-4)
    npt.assert_allclose(data['c'].mean(), 659.052e-6, atol=1e-9)
    npt.assert_allclose(data['q'].mean(), 9.56732e-3, atol=1e-7)
    npt.assert_allclose(data['T'].mean(), 28.5431 + 273.15, atol=1e-4)
    npt.assert_allclose(data['P'].mean(), 100.179e3, atol=1e0)

    toy_data = (
        'foobar baz\n'
        'asdf,0,2,3,4,5,6,7,9,0\n'
        'asdf,1,2,3,4,5,6,7,9,0\n'
        'asdf,2,2,3,4,5,6,7,9,1\n'
        'asdf,3,2,3,4,5,6,,9,0\n'
        'asdf,4,2,3,4,5,6,7,9,0\n'
        '# foo\n'
        'asdf,5,2,3,4,5,6,7,9,0\n'
        'asdf,6,2,3,4,5,6,7,xxx,0\n'
        'asdf,7,???,3,4,5,6,7,9,0\n'
        'asdf,8,2,3,4,5,6,7,9,0\n'
        'asdf,9,2,3,4,5,6,7,9,0\n'
        'asdf,10, 2,3,4,5,6,7,9,0\n'
        'asdf,11,-2,3,4,5,6,7,9,0\n')

    toy = HFData(
        io.BytesIO(toy_data.encode()),
        cols=(1, 2, 3, 7, 6, 4, 5),
        comments='#',
        skip_header=1,
        missing_values="???",
        converters={'q': _converter_func(10., 0)},
        flags=(9, 0),
        delimiter=",",
        rd_tol=0.1,
        ad_tol=2,
        bounds={'v': (0, np.inf)})

    npt.assert_allclose(toy.data_table['u'], [4, 5, 6])
    npt.assert_allclose(toy.data_table['v'], 3 * [2, ])
    npt.assert_allclose(toy.data_table['w'], 3 * [3, ])
    npt.assert_allclose(toy.data_table['q'], 3 * [70, ])
    npt.assert_allclose(toy.data_table['c'], 3 * [6, ])
    npt.assert_allclose(toy.data_table['T'], 3 * [4, ])
    npt.assert_allclose(toy.data_table['P'], 3 * [5, ])

if __name__ == '__main__':
    test_hfdata_from_txt()



"""Quick-and-dirty plots of :func:`~fluxpart.fluxpart.flux_partition` output"""

import matplotlib.pyplot as plt


PLTSTYLE = 'bmh'


def plot_hfdata(hfdata):
    """Note: converts q, c, T, and P from SI to common units"""
    plt.style.use(PLTSTYLE)
    fig, ax = plt.subplots(7, 1, figsize=(5, 8), sharex='all')
    ax[0] = _tsplot(hfdata['u'], r'u [m/s]', ax[0])
    ax[1] = _tsplot(hfdata['v'], r'v [m/s]', ax[1])
    ax[2] = _tsplot(hfdata['w'], r'w [m/s]', ax[2])
    ax[3] = _tsplot(1e3 * hfdata['q'], r'q [g/m^3]', ax[3])
    ax[4] = _tsplot(1e6 * hfdata['c'], r'c [mg/m^3]', ax[4])
    ax[5] = _tsplot(hfdata['T'] - 273.15, r'T [C]', ax[5])
    ax[6] = _tsplot(1e-3 * hfdata['P'], r'P [kPa]', ax[6])
    fig.tight_layout(pad=1, w_pad=0, h_pad=0)
    return fig


def _tsplot(series, yaxlab, ax=None):
    if ax is None:
        ax = plt.gca()
    ax.plot(series, linewidth=0.5)
    ax.set_ylabel('{} $\mathrm{{ {} }}$'.format(*yaxlab.split(maxsplit=1)))
    ax.set_xticklabels([])
    ax.locator_params(axis='y', tight=True, nbins=4)
    return ax


def plot_fluxes(fp_results, figtitle=None):
    """Note: converts Fq and Fc from SI to common units"""
    plt.style.use(PLTSTYLE)
    fcr = [1e6 * r['fluxes'].Fcr_mol for r in fp_results]
    fcp = [1e6 * r['fluxes'].Fcp_mol for r in fp_results]
    fc = [1e6 * r['fluxes'].Fc_mol for r in fp_results]
    fqe = [1e3 * r['fluxes'].Fqe_mol for r in fp_results]
    fqt = [1e3 * r['fluxes'].Fqt_mol for r in fp_results]
    fq = [1e3 * r['fluxes'].Fq_mol for r in fp_results]
    times = [r['label'] or i for i, r in enumerate(fp_results)]
    fig, (ax0, ax1) = plt.subplots(2, 1, figsize=(6, 6), sharex='all')
    ax0.plot(times, fc, label='Total')
    ax0.plot(times, fcr, 'o', label='Respiration')
    ax0.plot(times, fcp, '^', label='Photosynthesis')
    ax1.plot(times, fq, label='Total')
    ax1.plot(times, fqe, 'o', label='Evaporation')
    ax1.plot(times, fqt, '^', label='Transpiration')
    ax0.set_ylabel(r'CO2 $[\mathrm{umol/m^2/s}]$')
    ax1.set_ylabel(r'H2O $[\mathrm{mmol/m^2/s}]$')
    ax0.legend(loc='best', fontsize=10)
    ax1.legend(loc='best', fontsize=10)
    if figtitle:
        fig.suptitle(figtitle)
    plt.gcf().autofmt_xdate()
    fig.tight_layout()
    return fig


def plot_fluxes_mass(fp_results, figtitle=None):
    """Note: converts Fq and Fc from SI to common units"""
    plt.style.use(PLTSTYLE)
    fcr = [1e6 * r['fluxes'].Fcr for r in fp_results]
    fcp = [1e6 * r['fluxes'].Fcp for r in fp_results]
    fc = [1e6 * r['fluxes'].Fc for r in fp_results]
    fqe = [1e3 * r['fluxes'].Fqe for r in fp_results]
    fqt = [1e3 * r['fluxes'].Fqt for r in fp_results]
    fq = [1e3 * r['fluxes'].Fq for r in fp_results]
    times = [r['label'] or i for i, r in enumerate(fp_results)]
    fig, (ax0, ax1) = plt.subplots(2, 1, figsize=(6, 6), sharex='all')
    ax0.plot(times, fc, label='Total')
    ax0.plot(times, fcr, 'o', label='Respiration')
    ax0.plot(times, fcp, '^', label='Photosynthesis')
    ax1.plot(times, fq, label='Total')
    ax1.plot(times, fqe, 'o', label='Evaporation')
    ax1.plot(times, fqt, '^', label='Transpiration')
    ax0.set_ylabel(r'CO2 $[\mathrm{mg/m^2/s}]$')
    ax1.set_ylabel(r'H2O $[\mathrm{g/m^2/s}]$')
    ax0.legend(loc='best', fontsize=10)
    ax1.legend(loc='best', fontsize=10)
    if figtitle:
        fig.suptitle(figtitle)
    plt.gcf().autofmt_xdate()
    fig.tight_layout()
    return fig



import io
import numpy as np
import numpy.testing as npt
from fluxpart.util import stats2


def test_stats2():
    """Test stats2 func from fluxpart.util"""

    data = "7 8 4\n6 1 3\n10 6 6\n6 7 3\n8 2 4"
    dtype = [('v0', int), ('v1', int), ('v2', int)]
    arr = np.genfromtxt(io.BytesIO(data.encode()), dtype=dtype)
    ans = stats2(arr)

    npt.assert_allclose(ans.ave_v0, 37 / 5)
    npt.assert_allclose(ans.ave_v1, 24 / 5)
    npt.assert_allclose(ans.ave_v2, 4)
    npt.assert_allclose(ans.var_v0, 14 / 5)
    npt.assert_allclose(ans.var_v1, 97 / 10)
    npt.assert_allclose(ans.var_v2, 3 / 2)
    npt.assert_allclose(ans.cov_v0_v1, 3 / 5)
    npt.assert_allclose(ans.cov_v0_v2, 2)
    npt.assert_allclose(ans.cov_v1_v0, ans.cov_v0_v1)
    npt.assert_allclose(ans.cov_v1_v2, 1)
    npt.assert_allclose(ans.cov_v2_v0, ans.cov_v0_v2)
    npt.assert_allclose(ans.cov_v2_v1, ans.cov_v1_v2)

    data = "7 8 4\n6 1 3\n10 6 6\n6 7 3\n8 2 4"
    dtype = [('v0', int), ('v1', int), ('v2', int)]
    arr = np.genfromtxt(io.BytesIO(data.encode()), dtype=dtype)
    ans = stats2(arr, names=('v0', 'v2'))

    npt.assert_allclose(ans.ave_v0, 37 / 5)
    npt.assert_allclose(ans.ave_v2, 4)
    npt.assert_allclose(ans.var_v0, 14 / 5)
    npt.assert_allclose(ans.var_v2, 3 / 2)
    npt.assert_allclose(ans.cov_v0_v2, 2)
    npt.assert_allclose(ans.cov_v2_v0, ans.cov_v0_v2)

    assert not hasattr(ans, 'ave_v1')
    assert not hasattr(ans, 'var_v1')
    assert not hasattr(ans, 'cov_v0_v1')
    assert not hasattr(ans, 'cov_v1_v0')
    assert not hasattr(ans, 'cov_v1_v2')
    assert not hasattr(ans, 'cov_v2_v1')

if __name__ == "__main__":
    test_stats2()




import math
from types import SimpleNamespace
import numpy as np
import fluxpart.util as util
from fluxpart.containers import RootSoln, FluxComponents, QCData


def partition_from_wqc_series(w, q, c, wue, adjust_fluxes=True):
    """Partition H2O and CO2 fluxes using series data for w, q, and c.

    If a valid partitioning solution is not found for the passed series
    data, low-frequency (large-scale) components are progressively
    removed from the data until either a valid solution is found or
    the series decomposition is exhausted.

    Parameters
    ----------
    w,q,c : array_like
        1D time series for vertical wind speed `w` (m/s), water vapor
        concentration `q` (kg/m^3), and CO2 concentration `c` (kg/m^3).
    wue : float
        leaf-level water use efficiency, `wue` < 0, kg CO2 / kg H2O.
    adjust_fluxes : bool, optional
        Indicates whether the obtained partitioned fluxes should be
        adjusted so that the totals match the original data. Default is
        `adjust_fluxes` = True.

    Returns
    -------
    dict
        {'valid_partition': bool, 'partmssg': str,
        'fluxcomps': :class:`~fluxpart.containers.FluxComponents`,
        'rootsoln': :class:`~fluxpart.containers.RootSoln`,
        'qcdata': :class:`~fluxpart.containers.QCData`}

    Notes
    -----
    If a valid partitioning is not found, the returned `numersoln` and `qcdata`
    correspond to the final iteration attempted.

    """

    max_decomp_lvl = int(np.log2(w.size))
    wq_tot = np.cov((w, q))[0, 1]
    wc_tot = np.cov((w, c))[0, 1]

    # The loop progressively filters the data until a physically valid
    # partitioning is found or the loop/filter is exhausted. The first
    # iteration of progressive_lowcut removes only the mean value
    # (q'=q-<q>, etc.), so the first iteration uses the unfiltered
    # deviations.

    for cnt, lowcut_wqc in enumerate(progressive_lowcut(w, q, c)):
        cov = np.cov(lowcut_wqc)
        qcdata = QCData(
            wq=cov[0, 1],
            wc=cov[0, 2],
            var_q=cov[1, 1],
            var_c=cov[2, 2],
            corr_qc=cov[1, 2] / math.sqrt(cov[1, 1] * cov[2, 2]),
            wave_lvl=(max_decomp_lvl - cnt, max_decomp_lvl))

        rsoln, fcomp = partition_from_qc_averages(qcdata, wue)

        if rsoln.validroot:

            if adjust_fluxes:
                fcomp = adjust_partitioned_fluxes(fcomp, wue, wq_tot, wc_tot)

            valid_partition, partition_mssg = isvalid_partition(fcomp)

            if valid_partition:
                break

    if not rsoln.validroot:
        valid_partition = False
        partition_mssg = rsoln.validmssg
    if not valid_partition:
        fcomp = FluxComponents(*np.full(6, np.nan))

    return {'valid_partition': valid_partition,
            'partmssg': partition_mssg,
            'fluxcomps': fcomp,
            'rootsoln': rsoln,
            'qcdata': qcdata}


def partition_from_qc_averages(qcdata, wue):
    """Partition H2O and CO2 fluxes using interval average q and c data.

    All arguments are passed directly to the findroot function

    Parameters
    ----------
    qcdata : QCData namedtuple or equivalent namespace
    wue : float, kg CO2 / kg H2O
        Leaf-level water use efficiency, `wue` < 0

    Returns
    -------
    namedtuples
        :class:`~fluxpart.containers.RootSoln`,
        :class:`~fluxpart.containers.FluxComponents`

    """

    rsoln = findroot(qcdata, wue)
    if rsoln.validroot:
        fluxes = flux_components(rsoln.var_cp, rsoln.corr_cp_cr, qcdata, wue,
                                 rsoln.co2soln_id)
    else:
        fluxes = FluxComponents(*np.full(6, np.nan))
    return rsoln, fluxes


def findroot(qcdata, wue):
    """Calculate (corr_cp_cr, var_cp).

    Parameters
    ----------
    qcdata : namedtuple or equivalent namespace
        :class:`~fluxpart.containers.QCData`
    wue : float
        Leaf-level water use efficiency, `wue` < 0, kg CO2 / kg H2O.

    Returns
    -------
    namedtuple
        :class:`~fluxpart.containers.RootSoln`

    """


    # scale dimensional parameters so they have comparable magnitudes
    # H20: kg - > g
    # CO2: kg - > mg

    var_q = qcdata.var_q * 1e6
    var_c = qcdata.var_c * 1e12
    wq = qcdata.wq * 1e3
    wc = qcdata.wc * 1e6
    corr_qc = qcdata.corr_qc
    wue = wue * 1e3

    sd_q, sd_c = math.sqrt(var_q), math.sqrt(var_c)
    num0 = (corr_qc**2 - 1) * var_c * var_q * wue**2
    num1 = -2 * corr_qc * sd_c * sd_q * wq * wc  + var_c * wq**2 + var_q * wc**2
    numer = -num0 * num1
    denom = (-corr_qc * sd_c * sd_q * (wc + wq * wue) + var_c * wq + var_q * wc * wue)**2
    var_cp = numer / denom

    numer = - (corr_qc**2 - 1) * var_c * var_q * (wc - wq * wue)**2
    den0 = -2 * corr_qc * sd_c * sd_q * wue + var_c + var_q * wue**2
    den1 = -2 * corr_qc * sd_c * sd_q * wc * wq + var_c * wq**2 + var_q * wc**2
    denom = den0 * den1
    rho_sq = numer / denom
    corr_cp_cr = -math.sqrt(rho_sq)

    valid_root, valid_root_mssg = isvalid_root(corr_cp_cr, var_cp)

    co2soln_id = None
    sig_cr = np.nan
    if valid_root:
        valid_root = False
        valid_root_mssg = 'trial root did not satisfy equations'
        scaled_qcdata = QCData(
            wq=wq,
            wc=wc,
            var_q=var_q,
            var_c=var_c,
            corr_qc=corr_qc,
            wave_lvl=None)

        tol = 1e-12
        r0 = residual_func((corr_cp_cr, var_cp), scaled_qcdata, wue, 0)
        r1 = residual_func((corr_cp_cr, var_cp), scaled_qcdata, wue, 1)

        if abs(r0[0]) < tol and abs(r0[1]) < tol:
            co2soln_id = 0
            valid_root = True
            valid_root_mssg = ""
        if abs(r1[0]) < tol and abs(r1[1]) < tol:
            assert not co2soln_id
            co2soln_id = 1
            valid_root = True
            valid_root_mssg = ""

        if valid_root:
            qcdata = SimpleNamespace(
                var_q=var_q,
                var_c=var_c,
                wq=wq,
                wc=wc,
                corr_qc=corr_qc)

            wcr_ov_wcp = flux_ratio(var_cp, corr_cp_cr, qcdata, 'co2',
                                    co2soln_id)
            sig_cr = wcr_ov_wcp * math.sqrt(var_cp) / corr_cp_cr

    # re-scale dimensional variables to SI units
    return RootSoln(
        corr_cp_cr=corr_cp_cr,
        var_cp=var_cp * 1e-12,
        sig_cr=sig_cr * 1e-6,
        co2soln_id=co2soln_id,
        validroot=valid_root,
        validmssg=valid_root_mssg)


def flux_ratio(var_cp, corr_cp_cr, qcdata, ftype, farg):
    """Compute the nonstomatal:stomatal ratio of the H2O or CO2 flux.

    The ratio (either wqe/wqt or wcr/wcp) is found by solving Eq. 13
    of [SS08]_.

    Parameters
    ---------
    qcdata : namedtuple or equivalent namespace
        :class:`~fluxpart.containers.QCData`
    ftype : {'co2', 'h2o'}
        Specifies whether the flux is CO2 or H2O
    farg : number
        If `ftype` = 'co2', then `farg` = {0 or 1} specifies the root
        of Eq. 13b to be used to calculate the CO2 flux ratio wcr/wcp:
        `farg`=1 uses the '+' solution, `farg`=0 uses the '-' solution.
        If `ftype` = 'h2o', then `farg` is a float equal to the water
        use efficiency (wue < 0), kg/kg.

    Returns
    -------
    fratio : float or np.nan
        The requested flux ratio, wqe/wqt or wcr/wcp. Set to np.nan
        if solution is not real.

    Notes
    -----
    When solving for wqe/wqt, the '-' solution of the quadratic Eq. 13a
    is not relevant because only the '+' solution yields wqe/wqt > 0,
    which is required/assumed by the physical model in [SS08]_.

    """

    if ftype == 'co2' or ftype == 'CO2':
        sign = 1 if farg == 1 else -1
        num = qcdata.var_c
    elif ftype == 'h2o' or ftype == 'H2O':
        sign = 1
        num = farg**2 * qcdata.var_q
    else:
        raise ValueError("ftype must be 'co2' or 'h2o'")
    disc = 1 - 1 / corr_cp_cr**2 + num / var_cp / corr_cp_cr**2
    if disc < 0:
        fratio = np.nan
    else:
        fratio = corr_cp_cr**2 * (sign * math.sqrt(disc) - 1)
    return fratio


def flux_components(var_cp, corr_cp_cr, qcdata, wue, co2soln_id):
    """Calculate flux components for given (var_cp, corr_cp_cr) pair."""
    wcr_ov_wcp = flux_ratio(var_cp, corr_cp_cr, qcdata, 'co2', co2soln_id)
    # TODO: handle wcr_ov_wcp ~ -1
    wcp = qcdata.wc / (wcr_ov_wcp + 1)
    wcr = qcdata.wc - wcp
    wqt = wcp / wue
    wqe = qcdata.wq - wqt
    return FluxComponents(wq=qcdata.wq, wqt=wqt, wqe=wqe,
                          wc=qcdata.wc, wcp=wcp, wcr=wcr)


def residual_func(x, qcdata, wue, co2soln_id):
    """Residual function used with root finding routine.

    The two components of the residual are for Eqs. 15 and 18 of [SS08]_.

    """

    corr_cp_cr, var_cp = x
    wcr_ov_wcp = flux_ratio(var_cp, corr_cp_cr, qcdata, 'co2', co2soln_id)
    wqe_ov_wqt = flux_ratio(var_cp, corr_cp_cr, qcdata, 'h2o', wue)

    # Eq. 3, "f1"
    lhs = wue * qcdata.wq / qcdata.wc * (wcr_ov_wcp + 1)
    rhs = wqe_ov_wqt + 1
    resid1 = lhs - rhs

    # Eq. 4, "f2"
    lhs = wue * qcdata.corr_qc * math.sqrt(qcdata.var_c * qcdata.var_q) / var_cp
    rhs = (
        1 + wqe_ov_wqt + wcr_ov_wcp + wqe_ov_wqt * wcr_ov_wcp / corr_cp_cr**2)
    resid2 = lhs - rhs
    return [resid1, resid2]


def isvalid_root(corr_cp_cr, var_cp):
    isvalid = True
    mssg = ''
    if var_cp <= 0:
        isvalid = False
        mssg += 'var_cp <= 0; '
    if not -1 < corr_cp_cr < 0:
        isvalid = False
        mssg += 'corr_cp_cr <-1 OR >0; '
    # TODO: could add other bound checks?
    return isvalid, mssg


def isvalid_partition(pf):
    """Test if partitioned flux components are valid.

    Parameters
    ----------
    pf : namedtuple
        :class:`~fluxpart.containers.FluxComponents`

    """
    isvalid = True
    mssg = ''
    if pf.wqt <= 0:
        isvalid = False
        mssg += 'Fqt <= 0; '
    if pf.wqe <= 0:
        isvalid = False
        mssg += 'Fqe <= 0; '
    if pf.wcp >= 0:
        isvalid = False
        mssg += 'Fcp >= 0; '
    if pf.wcr <= 0:
        isvalid = False
        mssg += 'Fcr <= 0; '
    return isvalid, mssg


def adjust_partitioned_fluxes(fc, wue, wq_tot, wc_tot):
    """Adjust partitioned fluxes so they match measured totals.

    If filtering has been applied to the series data, covariances in
    the filtered data may differ from those in the original data.
    Consequently, partitioned flux totals may not match exactly the
    total fluxes indicated by the original data. Here, partitioned fluxes
    are adjusted proportionally so that they match the totals in the
    original data.

    Parameters
    ----------
    fc : :class:`~fluxpart.containers.FluxComponents` or equivalent
        Container holding partitioned flux components, kg/m^2/s.
    wue : float
        Leaf-level water use efficiency (`wue` < 0), kg CO2 / kg H2O
    wq_tot, wc_tot : float
        Desired net total H2O (`wq_tot`) and CO2 (`wc_tot`) fluxes,
        kg/m^2/s.

    Returns
    -------
    namedtuple
        :class:`~fluxpart.containers.FluxComponents`

    """

    wq_diff = wq_tot - (fc.wqe + fc.wqt)
    wqe = fc.wqe + wq_diff * (fc.wqe / (fc.wqt + fc.wqe))
    wqt = wq_tot - wqe
    wcp = wue * wqt
    wcr = wc_tot - wcp
    return FluxComponents(wq=wq_tot, wqt=wqt, wqe=wqe,
                          wc=wc_tot, wcp=wcp, wcr=wcr)


def progressive_lowcut(wind, vapor, co2):
    """Apply progressive lowcut filter to wind, vapor, and CO2 series.

    Uses wavelet decompostion to yield a sequence of (w, q, c) series
    in which low frequency (large scale) components are progressively
    removed from w, q, c.

    Parameters
    ----------
    wind,vapor,co2 : array-like
        1D time series of vertical wind velocity (w), water vapor
        concentration (q), and carbon dioxide concentration (c).

    Yields
    ------
    (lc_w, lc_q, lc_c) : tuple of arrays
        Arrays lc_w, lc_q, and lc_c are low cut (high pass) filtered
        versions of the passed w,q,c series.

    Notes
    -----
    The time series data are truncated at the maxium dyadic length
    (power of 2) before the filter is applied.

    """

    max_dyadic_len = 2**int(np.log2(np.asarray(co2).shape[0]))
    trunc_w = np.asarray(wind)[:max_dyadic_len]
    trunc_q = np.asarray(vapor)[:max_dyadic_len]
    trunc_c = np.asarray(co2)[:max_dyadic_len]
    lowcut_w = util.progressive_lowcut_series(trunc_w)
    lowcut_q = util.progressive_lowcut_series(trunc_q)
    lowcut_c = util.progressive_lowcut_series(trunc_c)
    for lowcut_series in zip(lowcut_w, lowcut_q, lowcut_c):
        yield lowcut_series


if __name__ == '__main__':
    pass



import os
from setuptools import setup

HERE = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(HERE, "fluxpart", "__version__.py")) as f:
    exec(f.read(), about)

LONG = ("Python 3 module implementing the Scanlon and Sahu (2008) procedure "
        "for partitioning eddy covariance measurements of water vapor and "
        "carbon dioxide fluxes into stomatal (transpiration, photosynthesis) "
        "and nonstomatal (evaporation, respiration) components.")

SHORT = ("Module for partitioning eddy covariance flux measurements.")

setup(
    name='fluxpart',
    version=about['__version__'],
    description=SHORT,
    long_description=LONG,
    url='https://github.com/usda-ars-ussl/fluxpart',
    author='Todd Skaggs',
    author_email='todd.skaggs@ars.usda.gov',
    license='Public Domain CC0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
    ],
    packages=['fluxpart'],
    install_requires=['numpy', 'matplotlib', 'pywavelets'],
    zip_safe=False,
    tests_require=['numpy'],
    test_suite='tests',
    extras_require={
        'tests': ['pytest'],
        'docs': ['sphinx >= 1.4']}
)
